"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.connectDocument = void 0;
const protocol_definitions_1 = require("@fluidframework/protocol-definitions");
const server_services_client_1 = require("@fluidframework/server-services-client");
const server_services_core_1 = require("@fluidframework/server-services-core");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const json_stringify_safe_1 = __importDefault(require("json-stringify-safe"));
const utils_1 = require("../utils");
const utils_2 = require("./utils");
const trace_1 = require("./trace");
const protocol_1 = require("./protocol");
const throttleAndUsage_1 = require("./throttleAndUsage");
const SummarizerClientType = "summarizer";
var ConnectDocumentStage;
(function (ConnectDocumentStage) {
    ConnectDocumentStage["ConnectDocumentStarted"] = "ConnectDocumentStarted";
    ConnectDocumentStage["VersionsChecked"] = "VersionsChecked";
    ConnectDocumentStage["ThrottleChecked"] = "ThrottleChecked";
    ConnectDocumentStage["TokenVerified"] = "TokenVerified";
    ConnectDocumentStage["RoomJoined"] = "RoomJoined";
    ConnectDocumentStage["ClientsRetrieved"] = "ClientsRetrieved";
    ConnectDocumentStage["MessageClientCreated"] = "MessageClientCreated";
    ConnectDocumentStage["MessageClientAdded"] = "MessageClientAdded";
    ConnectDocumentStage["TokenExpirySet"] = "TokenExpirySet";
    ConnectDocumentStage["MessageClientConnected"] = "MessageClientConnected";
    ConnectDocumentStage["SocketTrackerAppended"] = "SocketTrackerAppended";
    ConnectDocumentStage["SignalListenerSetUp"] = "SignalListenerSetUp";
    ConnectDocumentStage["JoinOpEmitted"] = "JoinOpEmitted";
})(ConnectDocumentStage || (ConnectDocumentStage = {}));
function composeConnectedMessage(messageSize, mode, serviceConfigurationBlockSize, serviceConfigurationMaxMessageSize, clientId, claims, version, clients) {
    const connectedMessage = {
        claims,
        clientId,
        existing: true,
        maxMessageSize: messageSize,
        mode,
        serviceConfiguration: {
            blockSize: serviceConfigurationBlockSize,
            maxMessageSize: serviceConfigurationMaxMessageSize,
        },
        initialClients: clients,
        initialMessages: [],
        initialSignals: [],
        supportedVersions: protocol_1.ProtocolVersions,
        supportedFeatures: {
            submit_signals_v2: true,
        },
        version,
    };
    return connectedMessage;
}
async function connectOrderer(socket, lumberjackProperties, lambdaDependencies, lambdaSettings, lambdaConnectionStateTrackers, tenantId, documentId, messageClient, startTime, clientId, claims, version, clients) {
    const { ordererManager, logger } = lambdaDependencies;
    const { numberOfMessagesPerTrace } = lambdaSettings;
    const { expirationTimer, connectionsMap } = lambdaConnectionStateTrackers;
    const connectDocumentOrdererConnectionMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.ConnectDocumentOrdererConnection, lumberjackProperties);
    const orderer = await ordererManager.getOrderer(tenantId, documentId).catch(async (err) => {
        const errMsg = `Failed to get orderer manager. Error: ${(0, json_stringify_safe_1.default)(err, undefined, 2)}`;
        connectDocumentOrdererConnectionMetric.error("Failed to get orderer manager", err);
        throw (0, utils_2.handleServerErrorAndConvertToNetworkError)(logger, errMsg, documentId, tenantId, err);
    });
    const connection = await orderer.connect(socket, clientId, messageClient).catch(async (err) => {
        const errMsg = `Failed to connect to orderer. Error: ${(0, json_stringify_safe_1.default)(err, undefined, 2)}`;
        connectDocumentOrdererConnectionMetric.error("Failed to connect to orderer", err);
        throw (0, utils_2.handleServerErrorAndConvertToNetworkError)(logger, errMsg, documentId, tenantId, err);
    });
    // Eventually we will send disconnect reason as headers to client.
    connection.once("error", (error) => {
        const messageMetaData = (0, utils_2.getMessageMetadata)(connection.documentId, connection.tenantId);
        logger.error(`Disconnecting socket on connection error: ${(0, json_stringify_safe_1.default)(error, undefined, 2)}`, { messageMetaData });
        server_services_telemetry_1.Lumberjack.error(`Disconnecting socket on connection error`, (0, server_services_telemetry_1.getLumberBaseProperties)(connection.documentId, connection.tenantId), error);
        expirationTimer.clear();
        socket.disconnect(true);
    });
    let clientJoinMessageServerMetadata;
    if (server_services_core_1.DefaultServiceConfiguration.enableTraces && (0, trace_1.sampleMessages)(numberOfMessagesPerTrace)) {
        clientJoinMessageServerMetadata = {
            connectDocumentStartTime: startTime,
        };
    }
    connection.connect(clientJoinMessageServerMetadata).catch(async (err) => {
        const errMsg = `Failed to connect to the orderer connection. Error: ${(0, json_stringify_safe_1.default)(err, undefined, 2)}`;
        connectDocumentOrdererConnectionMetric.error("Failed to establish orderer connection", err);
        throw (0, utils_2.handleServerErrorAndConvertToNetworkError)(logger, errMsg, documentId, tenantId, err);
    });
    connectionsMap.set(clientId, connection);
    if (connectionsMap.size > 1) {
        server_services_telemetry_1.Lumberjack.info(`Same socket is having multiple connections, connection number=${connectionsMap.size}`, (0, server_services_telemetry_1.getLumberBaseProperties)(connection.documentId, connection.tenantId));
    }
    connectDocumentOrdererConnectionMetric.success("Successfully established orderer connection");
    return composeConnectedMessage(connection.maxMessageSize, "write", connection.serviceConfiguration.blockSize, connection.serviceConfiguration.maxMessageSize, clientId, claims, version, clients);
}
function trackSocket(socket, tenantId, documentId, claims, { socketTracker }) {
    // Track socket and tokens for this connection
    if (socketTracker && claims.jti) {
        socketTracker.addSocketForToken((0, server_services_core_1.createCompositeTokenId)(tenantId, documentId, claims.jti), socket);
    }
}
function checkThrottle(tenantId, { throttlers, logger }) {
    const throttleErrorPerCluster = (0, throttleAndUsage_1.checkThrottleAndUsage)(throttlers.connectionsPerCluster, (0, throttleAndUsage_1.getSocketConnectThrottleId)("connectDoc"), tenantId, logger);
    if (throttleErrorPerCluster) {
        // eslint-disable-next-line @typescript-eslint/no-throw-literal
        throw throttleErrorPerCluster;
    }
    const throttleErrorPerTenant = (0, throttleAndUsage_1.checkThrottleAndUsage)(throttlers.connectionsPerTenant, (0, throttleAndUsage_1.getSocketConnectThrottleId)(tenantId), tenantId, logger);
    if (throttleErrorPerTenant) {
        // eslint-disable-next-line @typescript-eslint/no-throw-literal
        throw throttleErrorPerTenant;
    }
}
async function checkToken(token, tenantId, documentId, { tenantManager, revokedTokenChecker, logger }) {
    if (!token) {
        throw new server_services_client_1.NetworkError(403, "Must provide an authorization token");
    }
    const claims = (0, server_services_client_1.validateTokenClaims)(token, documentId, tenantId);
    try {
        if (revokedTokenChecker && claims.jti) {
            const isTokenRevoked = await revokedTokenChecker.isTokenRevoked(claims.tenantId, claims.documentId, claims.jti);
            if (isTokenRevoked) {
                throw new server_services_core_1.TokenRevokedError(403, "Permission denied. Token has been revoked", false /* canRetry */, true /* isFatal */);
            }
        }
        await tenantManager.verifyToken(claims.tenantId, token);
        return claims;
    }
    catch (error) {
        if ((0, server_services_client_1.isNetworkError)(error)) {
            throw error;
        }
        // We don't understand the error, so it is likely an internal service error.
        const errMsg = `Could not verify connect document token. Error: ${(0, json_stringify_safe_1.default)(error, undefined, 2)}`;
        throw (0, utils_2.handleServerErrorAndConvertToNetworkError)(logger, errMsg, claims.documentId, claims.tenantId, error);
    }
}
async function checkClusterDraining({ clusterDrainingChecker }, message, properties) {
    if (!clusterDrainingChecker) {
        return;
    }
    let clusterInDraining = false;
    try {
        clusterInDraining = await clusterDrainingChecker.isClusterDraining();
    }
    catch (error) {
        server_services_telemetry_1.Lumberjack.error("Failed to get cluster draining status. Will allow requests to proceed.", properties, error);
        clusterInDraining = false;
    }
    if (clusterInDraining) {
        // TODO: add a new error class
        server_services_telemetry_1.Lumberjack.info("Reject connect document request because cluster is draining.", {
            ...properties,
            tenantId: message.tenantId,
        });
        throw new server_services_client_1.NetworkError(503, "Cluster is not available. Please retry later.");
    }
}
async function joinRoomAndSubscribeToChannel(socket, tenantId, documentId, { logger }) {
    const clientId = (0, utils_1.generateClientId)();
    const room = {
        tenantId,
        documentId,
    };
    try {
        // Subscribe to channels.
        await Promise.all([
            socket.join((0, utils_2.getRoomId)(room)),
            socket.join((0, utils_2.getClientSpecificRoomId)(clientId)),
        ]);
        return [clientId, room];
    }
    catch (err) {
        const errMsg = `Could not subscribe to channels. Error: ${(0, json_stringify_safe_1.default)(err, undefined, 2)}`;
        throw (0, utils_2.handleServerErrorAndConvertToNetworkError)(logger, errMsg, documentId, tenantId, err);
    }
}
async function retrieveClients(tenantId, documentId, metricProperties, { clientManager, logger }, { maxNumberOfClientsPerDocument }) {
    const connectDocumentGetClientsMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.ConnectDocumentGetClients, metricProperties);
    const clients = await clientManager
        .getClients(tenantId, documentId)
        .then((response) => {
        connectDocumentGetClientsMetric.success("Successfully got clients from client manager");
        return response;
    })
        .catch(async (err) => {
        const errMsg = `Failed to get clients. Error: ${(0, json_stringify_safe_1.default)(err, undefined, 2)}`;
        connectDocumentGetClientsMetric.error("Failed to get clients during connectDocument", err);
        throw (0, utils_2.handleServerErrorAndConvertToNetworkError)(logger, errMsg, documentId, tenantId, err);
    });
    if (clients.length > maxNumberOfClientsPerDocument) {
        throw new server_services_client_1.NetworkError(429, "Too Many Clients Connected to Document", true /* canRetry */, false /* isFatal */, 5 * 60 * 1000 /* retryAfterMs (5 min) */);
    }
    return clients;
}
function createMessageClientAndJoinRoom(mode, client, claims, room, clientId, connectedTimestamp, supportedFeatures, { connectionTimeMap, scopeMap, roomMap, supportedFeaturesMap, }) {
    // Todo should all the client details come from the claims???
    // we are still trusting the users permissions and type here.
    const messageClient = client ?? {};
    messageClient.user = claims.user;
    messageClient.scopes = claims.scopes;
    messageClient.mode = (0, utils_2.isWriter)(claims.scopes, mode) ? "write" : "read";
    const isSummarizer = messageClient.details?.type === SummarizerClientType;
    // 1. Do not give SummaryWrite scope to clients that are not summarizers.
    // 2. Store connection timestamp for all clients but the summarizer.
    // Connection timestamp is used (inside socket disconnect event) to
    // calculate the client connection time (i.e. for billing).
    if (!isSummarizer) {
        messageClient.scopes = claims.scopes.filter((scope) => scope !== protocol_definitions_1.ScopeType.SummaryWrite);
        connectionTimeMap.set(clientId, connectedTimestamp);
    }
    // back-compat: remove cast to any once new definition of IClient comes through.
    messageClient.timestamp = connectedTimestamp;
    // Cache the scopes.
    scopeMap.set(clientId, messageClient.scopes);
    // Join the room to receive signals.
    roomMap.set(clientId, room);
    // Store the supported features for the client
    supportedFeaturesMap.set(clientId, supportedFeatures ?? {});
    return messageClient;
}
async function addMessageClientToClientManager(tenantId, documentId, clientId, messageClient, metricProperties, { clientManager, logger }) {
    const connectDocumentAddClientMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.ConnectDocumentAddClient, metricProperties);
    try {
        await clientManager.addClient(tenantId, documentId, clientId, messageClient);
        connectDocumentAddClientMetric.success("Successfully added client");
    }
    catch (err) {
        const errMsg = `Could not add client. Error: ${(0, json_stringify_safe_1.default)(err, undefined, 2)}`;
        connectDocumentAddClientMetric.error("Error adding client during connectDocument", err);
        throw (0, utils_2.handleServerErrorAndConvertToNetworkError)(logger, errMsg, documentId, tenantId, err);
    }
}
function setUpSignalListenerForRoomBroadcasting(socket, room, documentId, tenantId, { collaborationSessionEventEmitter, logger }) {
    collaborationSessionEventEmitter?.on("broadcastSignal", 
    // eslint-disable-next-line @typescript-eslint/no-misused-promises
    async (broadcastSignal) => {
        const { signalRoom, signalContent } = broadcastSignal;
        // No-op if the room (collab session) that signal came in from is different
        // than the current room. We reuse websockets so there could be multiple rooms
        // that we are sending the signal to, and we don't want to do that.
        if (signalRoom.documentId === room.documentId &&
            signalRoom.tenantId === room.tenantId) {
            try {
                const runtimeMessage = (0, utils_1.createRuntimeMessage)(signalContent);
                try {
                    socket.emitToRoom((0, utils_2.getRoomId)(signalRoom), "signal", runtimeMessage);
                }
                catch (error) {
                    const errorMsg = `Failed to broadcast signal from external API.`;
                    server_services_telemetry_1.Lumberjack.error(errorMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(signalRoom.documentId, signalRoom.tenantId), error);
                }
            }
            catch (error) {
                const errorMsg = `broadcast-signal content body is malformed`;
                throw (0, utils_2.handleServerErrorAndConvertToNetworkError)(logger, errorMsg, documentId, tenantId, error);
            }
        }
    });
}
async function connectDocument(socket, lambdaDependencies, lambdaSettings, lambdaConnectionStateTrackers, message, properties) {
    const { isTokenExpiryEnabled, maxTokenLifetimeSec } = lambdaSettings;
    const { expirationTimer } = lambdaConnectionStateTrackers;
    const connectionTrace = new trace_1.StageTrace(ConnectDocumentStage.ConnectDocumentStarted);
    const startTime = Date.now();
    const connectMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.ConnectDocument, properties);
    let tenantId = message.tenantId;
    let documentId = message.id;
    let uncaughtError;
    try {
        const [connectVersions, version] = (0, protocol_1.checkProtocolVersion)(message.versions);
        connectionTrace.stampStage(ConnectDocumentStage.VersionsChecked);
        checkThrottle(tenantId, lambdaDependencies);
        connectionTrace.stampStage(ConnectDocumentStage.ThrottleChecked);
        const claims = await checkToken(message.token, tenantId, documentId, lambdaDependencies);
        // check token validate tenantId/documentId for consistent, throw 403 if now.
        // Following change tenantId/documentId from claims, just in case future code changes that we can remember to use the ones from claim.
        tenantId = claims.tenantId;
        documentId = claims.documentId;
        connectionTrace.stampStage(ConnectDocumentStage.TokenVerified);
        await checkClusterDraining(lambdaDependencies, message, properties);
        const [clientId, room] = await joinRoomAndSubscribeToChannel(socket, tenantId, documentId, lambdaDependencies);
        connectionTrace.stampStage(ConnectDocumentStage.RoomJoined);
        const subMetricProperties = {
            ...(0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId),
            [server_services_telemetry_1.CommonProperties.clientId]: clientId,
        };
        const clients = await retrieveClients(tenantId, documentId, subMetricProperties, lambdaDependencies, lambdaSettings);
        connectionTrace.stampStage(ConnectDocumentStage.ClientsRetrieved);
        const connectedTimestamp = Date.now();
        const messageClient = createMessageClientAndJoinRoom(message.mode, message.client, claims, room, clientId, connectedTimestamp, message.supportedFeatures, lambdaConnectionStateTrackers);
        connectionTrace.stampStage(ConnectDocumentStage.MessageClientCreated);
        await addMessageClientToClientManager(tenantId, documentId, clientId, messageClient, subMetricProperties, lambdaDependencies);
        connectionTrace.stampStage(ConnectDocumentStage.MessageClientAdded);
        if (isTokenExpiryEnabled) {
            const lifeTimeMSec = (0, server_services_client_1.validateTokenClaimsExpiration)(claims, maxTokenLifetimeSec);
            expirationTimer.set(lifeTimeMSec);
        }
        connectionTrace.stampStage(ConnectDocumentStage.TokenExpirySet);
        const isWriterClient = (0, utils_2.isWriter)(messageClient.scopes ?? [], message.mode);
        connectMetric.setProperty("IsWriterClient", isWriterClient);
        const connectedMessage = isWriterClient
            ? await connectOrderer(socket, subMetricProperties, lambdaDependencies, lambdaSettings, lambdaConnectionStateTrackers, tenantId, documentId, messageClient, startTime, clientId, claims, version, clients)
            : composeConnectedMessage(1024 /* messageSize */, "read", server_services_core_1.DefaultServiceConfiguration.blockSize, server_services_core_1.DefaultServiceConfiguration.maxMessageSize, clientId, claims, version, clients);
        // back-compat: remove cast to any once new definition of IConnected comes through.
        connectedMessage.timestamp = connectedTimestamp;
        connectionTrace.stampStage(ConnectDocumentStage.MessageClientConnected);
        trackSocket(socket, tenantId, documentId, claims, lambdaDependencies);
        connectionTrace.stampStage(ConnectDocumentStage.SocketTrackerAppended);
        setUpSignalListenerForRoomBroadcasting(socket, room, documentId, tenantId, lambdaDependencies);
        connectionTrace.stampStage(ConnectDocumentStage.SignalListenerSetUp);
        const result = {
            connection: connectedMessage,
            connectVersions,
            details: messageClient,
        };
        socket.emitToRoom((0, utils_2.getRoomId)(room), "signal", (0, utils_1.createRoomJoinMessage)(result.connection.clientId, result.details));
        connectionTrace.stampStage(ConnectDocumentStage.JoinOpEmitted);
        connectMetric.setProperties({
            [server_services_telemetry_1.CommonProperties.clientId]: result.connection.clientId,
            [server_services_telemetry_1.CommonProperties.clientCount]: result.connection.initialClients.length + 1,
            [server_services_telemetry_1.CommonProperties.clientType]: result.details.details?.type,
        });
        return {
            connection: connectedMessage,
            connectVersions,
            details: messageClient,
        };
    }
    catch (err) {
        uncaughtError = err;
        throw err;
    }
    finally {
        connectMetric.setProperty("connectTrace", connectionTrace);
        if (!uncaughtError) {
            connectMetric.success(`Connect document successful`);
        }
        else {
            if (uncaughtError.code !== undefined) {
                connectMetric.setProperty(server_services_telemetry_1.CommonProperties.errorCode, uncaughtError.code);
            }
            connectMetric.error(`Connect document failed`, uncaughtError);
        }
    }
}
exports.connectDocument = connectDocument;
//# sourceMappingURL=connect.js.map