"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.disconnectDocument = void 0;
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const utils_1 = require("../utils");
const utils_2 = require("./utils");
const throttleAndUsage_1 = require("./throttleAndUsage");
async function disconnectDocument(socket, { clientManager, logger, throttleAndUsageStorageManager, socketTracker, }, { isClientConnectivityCountingEnabled }, { expirationTimer, connectionTimeMap, connectionsMap, roomMap, disconnectedClients, disconnectedOrdererConnections, }) {
    // Clear token expiration timer on disconnection
    expirationTimer.clear();
    const removeAndStoreP = [];
    // Send notification messages for all client IDs in the connection map
    for (const [clientId, connection] of connectionsMap) {
        if (disconnectedOrdererConnections.has(clientId)) {
            // We already removed this clientId once. Skip it.
            continue;
        }
        const messageMetaData = (0, utils_2.getMessageMetadata)(connection.documentId, connection.tenantId);
        logger.info(`Disconnect of ${clientId}`, { messageMetaData });
        server_services_telemetry_1.Lumberjack.info(`Disconnect of ${clientId}`, (0, server_services_telemetry_1.getLumberBaseProperties)(connection.documentId, connection.tenantId));
        connection
            .disconnect()
            .then(() => {
            // Keep track of disconnected clientIds so that we don't repeat the disconnect signal
            // for the same clientId if retrying when connectDocument completes after disconnectDocument.
            disconnectedOrdererConnections.add(clientId);
        })
            .catch((error) => {
            const errorMsg = `Failed to disconnect client ${clientId} from orderer connection.`;
            server_services_telemetry_1.Lumberjack.error(errorMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(connection.documentId, connection.tenantId), error);
        });
        if (isClientConnectivityCountingEnabled && throttleAndUsageStorageManager) {
            const connectionTimestamp = connectionTimeMap.get(clientId);
            if (connectionTimestamp) {
                removeAndStoreP.push((0, throttleAndUsage_1.storeClientConnectivityTime)(clientId, connection.documentId, connection.tenantId, connectionTimestamp, throttleAndUsageStorageManager));
            }
        }
    }
    // Send notification messages for all client IDs in the room map
    for (const [clientId, room] of roomMap) {
        if (disconnectedClients.has(clientId)) {
            // We already removed this clientId once. Skip it.
            continue;
        }
        const messageMetaData = (0, utils_2.getMessageMetadata)(room.documentId, room.tenantId);
        logger.info(`Disconnect of ${clientId} from room`, { messageMetaData });
        server_services_telemetry_1.Lumberjack.info(`Disconnect of ${clientId} from room`, (0, server_services_telemetry_1.getLumberBaseProperties)(room.documentId, room.tenantId));
        removeAndStoreP.push(clientManager
            .removeClient(room.tenantId, room.documentId, clientId)
            .then(() => {
            // Keep track of disconnected clientIds so that we don't repeat the disconnect signal
            // for the same clientId if retrying when connectDocument completes after disconnectDocument.
            disconnectedClients.add(clientId);
        })
            .catch((error) => {
            server_services_telemetry_1.Lumberjack.error(`Failed to remove client ${clientId} from client manager`, (0, server_services_telemetry_1.getLumberBaseProperties)(room.documentId, room.tenantId), error);
        }));
        try {
            socket.emitToRoom((0, utils_2.getRoomId)(room), "signal", (0, utils_1.createRoomLeaveMessage)(clientId));
        }
        catch (error) {
            const errorMsg = `Failed to emit signal to room ${clientId}, ${(0, utils_2.getRoomId)(room)}.`;
            server_services_telemetry_1.Lumberjack.error(errorMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(room.documentId, room.tenantId), error);
        }
    }
    // Clear socket tracker upon disconnection
    if (socketTracker) {
        socketTracker.removeSocket(socket.id);
    }
    await Promise.all(removeAndStoreP);
}
exports.disconnectDocument = disconnectDocument;
//# sourceMappingURL=disconnect.js.map