"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.checkThrottleAndUsage = exports.storeClientConnectivityTime = exports.getSubmitSignalThrottleId = exports.getSubmitOpThrottleId = exports.getSocketConnectThrottleId = void 0;
const server_services_core_1 = require("@fluidframework/server-services-core");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const json_stringify_safe_1 = __importDefault(require("json-stringify-safe"));
const getSocketConnectThrottleId = (tenantId) => `${tenantId}_OpenSocketConn`;
exports.getSocketConnectThrottleId = getSocketConnectThrottleId;
const getSubmitOpThrottleId = (clientId, tenantId) => `${clientId}_${tenantId}_SubmitOp`;
exports.getSubmitOpThrottleId = getSubmitOpThrottleId;
const getSubmitSignalThrottleId = (clientId, tenantId) => `${clientId}_${tenantId}_SubmitSignal`;
exports.getSubmitSignalThrottleId = getSubmitSignalThrottleId;
/**
 * Stores client connectivity time in a Redis list.
 */
async function storeClientConnectivityTime(clientId, documentId, tenantId, connectionTimestamp, throttleAndUsageStorageManager) {
    try {
        const now = Date.now();
        const connectionTimeInMinutes = (now - connectionTimestamp) / 60000;
        const storageId = server_services_core_1.clientConnectivityStorageId;
        const usageData = {
            value: connectionTimeInMinutes,
            tenantId,
            documentId,
            clientId,
            startTime: connectionTimestamp,
            endTime: now,
        };
        await throttleAndUsageStorageManager.setUsageData(storageId, usageData);
    }
    catch (error) {
        server_services_telemetry_1.Lumberjack.error(`ClientConnectivity data storage failed`, {
            [server_services_telemetry_1.CommonProperties.clientId]: clientId,
            [server_services_telemetry_1.BaseTelemetryProperties.tenantId]: tenantId,
            [server_services_telemetry_1.BaseTelemetryProperties.documentId]: documentId,
        }, error);
    }
}
exports.storeClientConnectivityTime = storeClientConnectivityTime;
/**
 * @returns ThrottlingError if throttled; undefined if not throttled or no throttler provided.
 */
function checkThrottleAndUsage(throttler, throttleId, tenantId, logger, usageStorageId, usageData, incrementWeight = 1) {
    if (!throttler) {
        return;
    }
    try {
        throttler.incrementCount(throttleId, incrementWeight, usageStorageId, usageData);
    }
    catch (error) {
        if (error instanceof server_services_core_1.ThrottlingError) {
            return error;
        }
        else {
            logger?.error(`Throttle increment failed: ${(0, json_stringify_safe_1.default)(error, undefined, 2)}`, {
                messageMetaData: {
                    key: throttleId,
                    eventName: "throttling",
                },
            });
            server_services_telemetry_1.Lumberjack.error(`Throttle increment failed`, {
                [server_services_telemetry_1.CommonProperties.telemetryGroupName]: "throttling",
                [server_services_telemetry_1.BaseTelemetryProperties.tenantId]: tenantId,
            }, error);
        }
    }
}
exports.checkThrottleAndUsage = checkThrottleAndUsage;
//# sourceMappingURL=throttleAndUsage.js.map